program WEinvertfix

  use sep
  use fop_mod
  use gradiometry_reg_mod
  use reg_laplac_mod
  use cgstep_dble_mod
  use solver_reg_mod 
  implicit none

  integer :: ii,jj,kk,ll
  integer :: n1,n2,n3
  real    :: d1,d2,d3
  real    :: o1,o2,o3
 
  real, allocatable :: dataot(:,:)
  real, allocatable :: data3(:,:,:)
  double precision, allocatable  :: Fop(:),m(:),d(:),lap(:,:),dtdt(:,:)
  
  integer :: niter
  real    :: eps
  logical :: verbose

  integer :: j1,j2

  call sep_init()
  
  call from_history('n1',n1,1 )
  call from_history('n2',n2,1 )
  call from_history('n3',n3,1 )
  
  call from_history('d1',d1,1.)
  call from_history('d2',d2,1.)
  call from_history('d3',d3,1.);! d3=1./3/1000.
  
  call from_history('o1',o1,0.)
  call from_history('o2',o2,0.)
  call from_history('o3',o3,0.)
  
  call from_param('j1',j1,1)
  call from_param('j2',j2,1)

  call to_history('n1',n1/j1)
  call to_history('d1',d1*j1)
  call to_history('n2',n2/j2)
  call to_history('d2',d2*j2)
  call to_history('n3',1 )
  call to_history('d3',1.)
  call to_history('o3',0.)
  call to_history('n4',1 )
  call to_history('d4',1.)
  call to_history('o4',0.)
  
  call from_param('niter',niter,50)
  call from_param('eps'  ,eps  ,1.)

  verbose=.true.
  call from_param('verb'   ,verbose,verbose)
  call from_param('verbose',verbose,verbose)
  
  if (exist_file("div")) then
    call to_history('n1',n1/j1,'div')
    call to_history('n2',n2/j2,'div')
    call to_history('o1',o1,'div')
    call to_history('o2',o2,'div')
    call to_history('d1',d1*j1,'div')
    call to_history('d2',d2*j2,'div')
    call to_history('n3',1 ,'div')
    call to_history('d3',1.,'div')
    call to_history('o3',0.,'div')
    call to_history('n4',1 ,'div')
    call to_history('d4',1.,'div')
    call to_history('o4',0.,'div')
  end if

  call sep_close()
  
  allocate(Fop(n1*n2),m(n1*n2),d(n1*n2))
  Fop=0.;m=0.;d=0.;dataot=0.
  
  ! Prepare Fop and d
  allocate(data3(n1,n2,3),lap(n1,n2),dtdt(n1,n2)); data3=0.
  call gradiometry_reg_init(n1_in=n1,n2_in=n2,d1_in=d1,d2_in=d2,dt_in=d3)
  
  ! Gradiometry operation
  call sreed('in',data3(:,:,1),4*n1*n2)
  data3(:,:,3)=data3(:,:,2)
  data3(:,:,2)=data3(:,:,1)
  call sreed('in',data3(:,:,1),4*n1*n2)
  do kk=2,n3-1
    data3(:,:,3)=data3(:,:,2)
    data3(:,:,2)=data3(:,:,1)
    call sreed('in',data3(:,:,1),4*n1*n2)
    
    call grad_laplacian_reg(data3(:,:,2),lap )
    call grad_deltdelt     (data3       ,dtdt)
    
    do jj=1,n2
    do ii=1,n1
      Fop((jj-1)*n1+ii)=Fop((jj-1)*n1+ii)+lap(ii,jj)*lap(ii,jj)
      d  ((jj-1)*n1+ii)=d  ((jj-1)*n1+ii)+lap(ii,jj)*dtdt(ii,jj)
    end do
    end do
    
  end do
  deallocate(data3,lap,dtdt)
  call gradiometry_reg_close()
  
  ! Initialize Fop and Aop
  call fop_init(Fop,n1*n2)
  call reg_laplac_init(n1,n2,d1,d2)
  
  ! Call solver
  call solver_reg(m,d,Fop=fop_lop,Aop=reg_laplac_lop,stepper=cgstep,nAop=n1*n2,niter=niter,eps0=dble(eps),verb=verbose)

  ! CLose stepper and operators
  call fop_close()
  call reg_laplac_close()
  call cgstep_close()

  ! Convert to Velocity
  allocate(dataot(n1,n2))
  do ii=1,n1
  do jj=1,n2
    dataot(ii,jj)=real(sqrt(m((jj-1)*n1+ii)))
  end do
  end do

  ! Write out model
  call srite('out',dataot,4*n1*n2)

  ! Compute explicit Inversion and write out
  if (exist_file("div")) then
    do ii=1,n1
    do jj=1,n2
      dataot(ii,jj)=real(sqrt(d((jj-1)*n1+ii)/Fop((jj-1)*n1+ii)))
    end do
    end do
    call srite('div',dataot,4*n1*n2)
  end if
  
  ! Deallocate, close and exit
  deallocate(Fop,m,d,dataot)
  call exit()
  
end program
