program RandBoundarySources

  ! Program to create positions and time series of random sources
  ! Define a plane-plane and a distribution around it

  ! From param:
  ! ns: number of sources (to be created or accepted)
  ! stype: source type

  ! bndx: Distance from z-edges
  ! bndx: Distance from x-edges
  ! cntt: center of t-distribution

  
  use sep
  implicit none
  
  real,allocatable  :: sources_zxt(:,:)

  integer           :: ns,stype,is

  real              :: cntt,delt
  integer           :: bndx,bndz

  real              :: cntxz,delxz,x,z
  
  integer           :: n1,n2
  real              :: d1,d2
  real              :: harvest_xz,harvest_t

  ! Header I/O
  call sep_init()
  
  call from_param("ns",ns,100)
  
  call from_param("bndz",bndz,90)
  call from_param("bndx",bndx,90)
  call from_param("delt",delt, 11.)
  call from_param("cntt",cntt, 11.)

  call from_param("stype",stype,1)

  call from_param("n1",n1,500)  ! Maximum model sizes
  call from_param("n2",n2,500)  ! Maximum model sizes
  call from_param("d1",d1,1.)  
  call from_param("d2",d2,1.)  

  call to_history("n1",ns)
  call to_history("n2", 4)
  call to_history("d1",1.)
  call to_history("d2",1.)
  call to_history("o1",1.)
  call to_history("o2",1.)

  call sep_close()
  
  ! Compute center and its maximum distance from the center of the boundary contour
  delxz=2*(n1-bndx-bndx)+2*(n2-bndz-bndz);
  cntxz=delxz/2
  delxz=delxz/2
  

  ! Allocate
  allocate(sources_zxt(ns,4))
  
  ! Find accepted sources
  is=0
  do while (is.lt.ns)
    
    ! Get random numbers
    call random_number(harvest_t);   harvest_t  = cntt  + (harvest_t -.5)*2*delt
    call random_number(harvest_xz);  harvest_xz = cntxz + (harvest_xz-.5)*2*delxz
    
    x=bndx; z=bndz+harvest_xz-(n1-bndx-bndx)-(n2-bndz-bndz)-(n1-bndx-bndx)
    if (harvest_xz.lt.(n1-bndx-bndx)+(n2-bndz-bndz)+(n1-bndx-bndx)) then
      x=bndx+harvest_xz-(n1-bndx-bndx)-(n2-bndz-bndz); z=n2-bndz
    end if
    if (harvest_xz.lt.(n1-bndx-bndx)+(n2-bndz-bndz)               ) then
      x=n1-bndx; z=bndz+harvest_xz-(n1-bndz-bndz);
    end if 
    if (harvest_xz.lt.(n1-bndx-bndx)                              ) then
       x=bndx+harvest_xz; z=bndz
    end if

    is = is+1
    sources_zxt(is,:) = (/ x*d1,z*d2,harvest_t /)

  end do

  ! Source type
  sources_zxt(:,4) = stype

  ! Write out
  call srite("out",sources_zxt,4*size(sources_zxt))

  deallocate(sources_zxt)
  
end program
