program RandSources

  ! Program to create positions and time series of random sources
  ! Define a plane-plane and a distribution around it

  ! From param:
  ! ns: number of sources (to be created or accepted)
  ! stype: source type

  ! cntz: center of z-distribution
  ! cntx: center of x-distribution
  ! cntt: center of t-distribution

  ! cntz: max deveation from z-center
  ! cntx: max deveation from x-center
  ! cntt: max deveation from t-center

  ! distz: distribution type in z-direction ('flat','norm')
  ! distx: distribution type in x-direction ('flat','norm')
  ! distt: distribution type in t-direction ('flat','norm')

  ! dz = maximum distance from plane
  ! gz = std-deviation of normal distribution in z-direction
  ! gx = std-deviation of normal distribution in y-direction
  ! gt = std-deviation of normal distribution in t-direction
  
  use sep
  implicit none
  
  real,allocatable  ::  sources_zxt(:,:)

  integer           :: ns,stype,is

  real              :: cntx,cntz,cntt
  real              :: delx,delz,delt
  
  real              :: gx,gz,gt
  integer           :: distz,distx,distt
  real              :: harvest_x,harvest_z,harvest_t,harvest_e,expectency

  ! Header I/O
  call sep_init()
  
  call from_param("ns",ns,100)
  
  call from_param("delz",delz,10.)
  call from_param("delx",delx,10.)
  call from_param("delt",delt,10.)
  
  call from_param("cntz",cntz,275.)
  call from_param("cntx",cntx,600.)
  call from_param("cntt",cntt, 11.)
  
  call from_param("gz",gz,100.)
  call from_param("gx",gx,100.)
  call from_param("gt",gt, 10.)
  
  call from_param("distz",distz,1)
  call from_param("distx",distx,1)
  call from_param("distt",distt,1)
  
  call from_param("stype",stype,1)
  
  call to_history("n1",ns)
  call to_history("n2", 4)
  call to_history("d1",1.)
  call to_history("d2",1.)
  call to_history("o1",1.)
  call to_history("o2",1.)

  call sep_close()
  
  ! Allocate
  allocate(sources_zxt(ns,4))
  
  ! Find accepted sources
  is=0
  do while (is.lt.ns)

    ! Get random numbers
    call random_number(harvest_z);  harvest_z = cntz + (harvest_z-.5)*2*delz
    call random_number(harvest_x);  harvest_x = cntx + (harvest_x-.5)*2*delx
    call random_number(harvest_t);  harvest_t = cntt + (harvest_t-.5)*2*delt
    call random_number(harvest_e);

    ! Build expentency according to requested distributions
    if (distt.eq.1) then

    if (distz.eq.1) then
    if (distx.eq.1) then
      expectency = 1.                                * 1.                                * 1.
    else
      expectency = exp(-(harvest_z-cntz)/(2*gz)**2.) * 1.                                * 1.
    end if
    else
    if (distx.eq.1) then
      expectency = 1.                                * exp(-(harvest_x-cntx)/(2*gx)**2.) * 1.
    else
      expectency = exp(-(harvest_z-cntz)/(2*gz)**2.) * exp(-(harvest_x-cntx)/(2*gx)**2.) * 1.
    end if
    end if

    else

    if (distz.eq.1) then
    if (distx.eq.1) then
      expectency = 1.                                * exp(-(harvest_x-cntx)/(2*gx)**2.) * exp(-(harvest_t-cntt)/(2*gt)**2.)
    else
      expectency = exp(-(harvest_z-cntz)/(2*gz)**2.) * exp(-(harvest_x-cntx)/(2*gx)**2.) * exp(-(harvest_t-cntt)/(2*gt)**2.)
    end if
    else
    if (distx.eq.1) then
      expectency = 1.                                * 1.                                * exp(-(harvest_t-cntt)/(2*gt)**2.)
    else
      expectency = exp(-(harvest_z-cntz)/(2*gz)**2.) * 1.                                * exp(-(harvest_t-cntt)/(2*gt)**2.)
    end if
    end if

    end if
    
    ! If accepted, store z,x,t
    if (harvest_e.le.expectency) then
      is = is+1
      sources_zxt(is,:) = (/ harvest_z,harvest_x,harvest_t /)
    end if

  end do

  ! Source type
  sources_zxt(:,4) = stype

  ! Write out
  call srite("out",sources_zxt,4*size(sources_zxt))

  deallocate(sources_zxt)
  
end program
