program WEinvertfix

  use sep
  use fop_mod
  use reg_laplac_mod
  use cgstep_dble_mod  
  use solver_reg_mod 
  implicit none

  integer :: ii,jj,kk,ll
  integer :: n1,n2,n3
  real    :: d1,d2,d3
  real    :: o1,o2,o3

  real    :: dxdx,dydy,dtdt

  real, allocatable :: dataot(:,:)
  real, allocatable :: datain1(:,:),datain2(:,:),datain3(:,:)
  double precision, allocatable  :: Fop(:),m(:),d(:),div(:),mmov(:,:)

  integer :: niter
  real    :: eps
  logical :: verbose

  integer :: j1,j2

  call sep_init()
  
  call from_history('n1',n1,1 )
  call from_history('n2',n2,1 )
  call from_history('n3',n3,1 )
  
  call from_history('d1',d1,1.)
  call from_history('d2',d2,1.)
  call from_history('d3',d3,1.);! d3=1./3/1000.
  
  call from_history('o1',o1,0.)
  call from_history('o2',o2,0.)
  call from_history('o3',o3,0.)
  
  call from_param('j1',j1,1)
  call from_param('j2',j2,1)

  call to_history('n1',n1/j1)
  call to_history('d1',d1*j1)
  call to_history('n2',n2/j2)
  call to_history('d2',d2*j2)
  call to_history('n3',1 )
  call to_history('d3',1.)
  call to_history('o3',0.)
  call to_history('n4',1 )
  call to_history('d4',1.)
  call to_history('o4',0.)
  
  call from_param('niter',niter,50)
  call from_param('eps'  ,eps  ,1.)

  verbose=.true.
  call from_param('verb'   ,verbose,verbose)
  call from_param('verbose',verbose,verbose)
  
  if (exist_file("Fop")) then
    call to_history('n1',n1/j1,'Fop')
    call to_history('n2',n2/j2,'Fop')
    call to_history('o1',o1,'Fop')
    call to_history('o2',o2,'Fop')
    call to_history('d1',d1*j1,'Fop')
    call to_history('d2',d2*j2,'Fop')
    call to_history('n3',1 ,'Fop')
    call to_history('d3',1.,'Fop')
    call to_history('o3',0.,'Fop')
    call to_history('n4',1 ,'Fop')
    call to_history('d4',1.,'Fop')
    call to_history('o4',0.,'Fop')
  end if
  if (exist_file("d")) then
    call to_history('n1',n1/j1,'d')
    call to_history('n2',n2/j2,'d')
    call to_history('o1',o1,'d')
    call to_history('o2',o2,'d')
    call to_history('d1',d1*j1,'d')
    call to_history('d2',d2*j2,'d')
    call to_history('n3',1 ,'d')
    call to_history('d3',1.,'d')
    call to_history('o3',0.,'d')
    call to_history('n4',1 ,'d')
    call to_history('d4',1.,'d')
    call to_history('o4',0.,'d')
  end if
  if (exist_file("div")) then
    call to_history('n1',n1/j1,'div')
    call to_history('n2',n2/j2,'div')
    call to_history('o1',o1,'div')
    call to_history('o2',o2,'div')
    call to_history('d1',d1*j1,'div')
    call to_history('d2',d2*j2,'div')
    call to_history('n3',1 ,'div')
    call to_history('d3',1.,'div')
    call to_history('o3',0.,'div')
    call to_history('n4',1 ,'div')
    call to_history('d4',1.,'div')
    call to_history('o4',0.,'div')
  end if
  if (exist_file("mmov")) then
    call to_history('n1',n1/j1,'mmov')
    call to_history('n2',n2/j2,'mmov')
    call to_history('o1',o1,'mmov')
    call to_history('o2',o2,'mmov')
    call to_history('d1',d1*j1,'mmov')
    call to_history('d2',d2*j2,'mmov')
    call to_history('n3',niter,'mmov')
    call to_history('d3',1.,'mmov')
    call to_history('o3',0.,'mmov')
    call to_history('n4',1 ,'mmov')
    call to_history('d4',1.,'mmov')
    call to_history('o4',0.,'mmov')
  end if
  
  call sep_close()
  
  allocate(Fop(n1/j1*n2/j2),m(n1/j1*n2/j2),d(n1/j1*n2/j2),dataot(n1/j1,n2/j2),mmov(n1/j1*n2/j2,niter))
  Fop=0.;m=0.;d=0.;dataot=0.;mmov=0.
  
  ! Prepare Fop and d
  allocate(datain1(n1,n2),datain2(n1,n2),datain3(n1,n2))
  datain1=0.;datain2=0.;datain3=0.
  call sreed('in',datain1,4*n1*n2)
  datain3=datain2
  datain2=datain1
  call sreed('in',datain1,4*n1*n2)
  do kk=2,n3-1
    datain3=datain2
    datain2=datain1
    call sreed('in',datain1,4*n1*n2)
    
    do ii=2,n1/j1-1
    do jj=2,n2/j2-1
      dxdx = (datain2((ii-1)*j1+1+1,(jj-1)*j2+1  )-2*datain2((ii-1)*j1+1,(jj-1)*j2+1)+datain2((ii-1)*j1+1-1,(jj-1)*j2+1  ))/d1/d1
      dydy = (datain2((ii-1)*j1+1  ,(jj-1)*j2+1+1)-2*datain2((ii-1)*j1+1,(jj-1)*j2+1)+datain2((ii-1)*j1+1  ,(jj-1)*j2+1-1))/d2/d2
      dtdt = (datain1((ii-1)*j1+1  ,(jj-1)*j2+1  )-2*datain2((ii-1)*j1+1,(jj-1)*j2+1)+datain3((ii-1)*j1+1  ,(jj-1)*j2+1  ))/d3/d3
      ! if ((dxdx+dydy).ne.sign((dxdx+dydy),dtdt)) cycle
      Fop((jj-1)*n1/j1+ii)=Fop((jj-1)*n1/j1+ii)+dble(dxdx+dydy)*dble(dxdx+dydy)
      d  ((jj-1)*n1/j1+ii)=d  ((jj-1)*n1/j1+ii)+dble(dtdt     )*dble(dxdx+dydy)
    end do
    end do
    
  end do
  deallocate(datain1,datain2,datain3)

  n1=n1/j1
  n2=n2/j2
  
  jj=1
  do ii=2,n1-1
    Fop((jj-1)*n1+ii)=Fop((jj  )*n1+ii)
    d  ((jj-1)*n1+ii)=d  ((jj  )*n1+ii)
  end do
  jj=n2
  do ii=2,n1-1
    Fop((jj-1)*n1+ii)=Fop((jj-2)*n1+ii)
    d  ((jj-1)*n1+ii)=d  ((jj-2)*n1+ii)
  end do

  ii=1
  do jj=2,n2-1
    Fop((jj-1)*n1+ii)=Fop((jj-1)*n1+ii+1)
    d  ((jj-1)*n1+ii)=d  ((jj-1)*n1+ii+1)
  end do
  ii=n1
  do jj=2,n2-1
    Fop((jj-1)*n1+ii)=Fop((jj-1)*n1+ii-1)
    d  ((jj-1)*n1+ii)=d  ((jj-1)*n1+ii-1)
  end do

  Fop(( 1-1)*n1+ 1)=Fop(( 2-1)*n1+ 2  )
  Fop(( 1-1)*n1+n1)=Fop(( 2-1)*n1+n1-1)
  Fop((n2-1)*n1+ 1)=Fop((n2-1)*n1+2   )
  Fop((n2-1)*n1+n1)=Fop((n2-1)*n1+n1-1)
  d  (( 1-1)*n1+ 1)=d  (( 2-1)*n1+ 2  )
  d  (( 1-1)*n1+n1)=d  (( 2-1)*n1+n1-1)
  d  ((n2-1)*n1+ 1)=d  ((n2-1)*n1+2   )
  d  ((n2-1)*n1+n1)=d  ((n2-1)*n1+n1-1)
  
 ! write(0,*) 'Fop'
 ! write(0,*)  maxval(Fop),minval(Fop)
 ! write(0,*) 'd'
 ! write(0,*)  maxval(d),minval(d)
  if (exist_file("Fop")) then
    call srite('Fop',real(Fop),4*size(Fop))
  end if
  if (exist_file("d")) then
    call srite('d'  ,real(d  ),4*size(d  ))
  end if

  if (exist_file("div")) then
    allocate(div(n1*n2))
    do ii=1,n1
    do jj=1,n2
      div((jj-1)*n1+ii)=sqrt(d((jj-1)*n1+ii)/Fop((jj-1)*n1+ii))
    end do
    end do
    call srite('div',real(div),4*size(div))
    deallocate(div)
  end if

  ! Initialize Fop and Aop
  call fop_init(Fop,n1*n2)
  call reg_laplac_init(n1,n2,d1,d2)
  
  ! Call solver
  call solver_reg(m,d,Fop=fop_lop,Aop=reg_laplac_lop,stepper=cgstep,nAop=n1*n2,niter=niter,eps0=dble(eps),verb=verbose,mmov=mmov)

  ! CLose stepper and operators
  call fop_close()
  call reg_laplac_close()
  call cgstep_close()
  
! ! Compute the explicit inversion
!  do ii=1,n1
!  do jj=1,n2
!    m((jj-1)*n1+ii)=d((jj-1)*n1+ii)/Fop((jj-1)*n1+ii)
!  end do
!  end do

  do ii=1,n1
  do jj=1,n2
    dataot(ii,jj)=real(sqrt(m((jj-1)*n1+ii)))
  end do
  end do

  call srite('out',dataot,4*n1*n2)

  if (exist_file("mmov")) then
    mmov=sqrt(mmov)
    call srite('mmov',real(mmov),4*n1*n2*niter)
  end if

  deallocate(Fop,m,d,dataot,mmov)
  call exit()
end program
