program acoustic2dvd
  use sep
  use sources_mod
  use sep_timers_mod

  implicit none
  
  ! Grid parameters
  integer                           :: nx,nz,npad,nop
  real                              :: x0,z0
  integer                           :: it,iz,ix,iy,iop
  real                              :: T,tapco
  
  ! Computational time variables
  integer                           :: nt_comp,jt_verb
  real                              :: dz,dx,dt_comp
  real                              :: vmin,vmax

  ! Finite Difference variables
  real                              :: a1x,a2x,a3x,a4x,a5x,a1z,a2z,a3z,a4z,a5z
  real,allocatable                  :: u(:,:,:),ut(:,:)
  real,allocatable                  :: data(:,:)
  real,allocatable                  :: taper(:)

  ! Medium variables
  real,allocatable                  :: vv(:,:),rh(:,:),ik(:,:)
  real                              :: rho

  ! Source and receiver positions variables
  integer,allocatable               :: sources_zxt(:,:),receiver_zxt(:,:),us_zx(:,:),ur_zx(:,:)   
  real,allocatable                  :: sources_zxt_in(:,:),receiver_zxt_in(:,:),us(:,:),ur(:,:)
  integer                           :: ns,nsx,nsz,fsx,fsz,jsx,jsz,stype
  integer                           :: nr,nrt,nrx,nrz,frx,frz,jrx,jrz,rtype
  integer                           :: isx,isz,irx,irz,is,ir,is2,ns2,ir2,nr2
  integer                           :: jt_data,nt_data
  real                              :: dt_data

  ! For snap shots
  real,allocatable                  :: usnap(:,:,:)
  integer                           :: jt_snap,nt_snap
  real                              :: dt_snap

  ! Source parameters
  real,allocatable                  :: sources(:,:)
  real                              :: f0,w0,smth_fact
  integer                           :: nst,smth

  ! ???
  real                              :: id,in
  real                              :: dxi,dzi
  real                              :: dt2,rr,delp
  integer                           :: ixe,ize,jt

  real,parameter                    :: pi=3.1415926535

  logical                           :: snap, ssquare, freesurf

  ! For timers
  logical           :: logic
  integer           :: tm1,tm2,tm3

  write(0,*) '----------------------------'
  write(0,*) 'Variable Density Acoustic FD'
  logic = init_sep_timers()
  logic = setup_next_timer("Total----",tm1)
  logic = setup_next_timer("Overhead-",tm2)
  logic = setup_next_timer("Core-----",tm3)
  call start_timer_num(tm1)
  call start_timer_num(tm2)
  
  call sep_init()


  ! *********************************************************************
  !  Setup Computational Grid and Medium Parameters
  ! *********************************************************************
  write(0,*) '----------------------------'
  write(0,*) 'Setup Computational Grid'
  
  nop=5                            ! Operator padding zone
  call from_param("npad",npad,20)  ! Padding zone
  call from_param("tapco",tapco,0.0002)
  call from_param("T",T)           ! Total recording time
  call from_param("freesurf",freesurf,.true.) ! Free (top) surface or not.
  

  ! Geometry of the experiments (read from prepared file)
  call from_history("n1",nz)
  call from_history("n2",nx)
  call from_history("d1",dz)
  call from_history("d2",dx)
  call from_history("o1",z0)
  call from_history("o2",x0)

  write(0,*) "nz,nx",nz,nx

  !!! Read Medium Parameters
  allocate(vv(1-nop:nz+nop,1-nop:nx+nop)); vv = 0.
  allocate(rh(1-nop:nz+nop,1-nop:nx+nop)); rh = 0.
  allocate(ik(1-nop:nz+nop,1-nop:nx+nop)); ik = 0.

  call sreed("in",vv(1:nz,1:nx),4*nz*nx)
  call sreed("in",rh(1:nz,1:nx),4*nz*nx)
  call sreed("in",ik(1:nz,1:nx),4*nz*nx)

  ! Invert incompressibility
  do ix=1,nx
    do iz=1,nz
      ik(iz,ix) = 1./dble(ik(iz,ix))
    end do
  end do
  ! Extend medium parameters ???
  rh(0,:) = .5*rh(1,:)
  
  !!! Determine timesteps
  vmax = 0.
  vmin = vv(1,1)
  do ix=1,nx;  do iz=1,nz
        vmax=max(vmax,vv(iz,ix))
        vmin=min(vmin,vv(iz,ix))
  end do    ;  end do
  dt_comp = .333*min(dz,dx)/vmax;
  nt_comp = T/dt_comp
  jt_verb = int(1./dt_comp)
  
  
  ! *********************************************************************
  !  Source/Receicer Section
  ! *********************************************************************
  write(0,*) '----------------------------'
  write(0,*) 'Source/Receiver Setup'
  
  !! Receiver Output
  call from_param("nrt",nrt,-1)
  call from_param("nrz",nrz,1 )
  call from_param("nrx",nrx,1 )
  call from_param("frz",frz,nz/2+1)
  call from_param("frx",frx,nx/2+1)
  call from_param("jrz",jrz,1 )
  call from_param("jrx",jrx,1 )
  call from_param("rtype",rtype,1)
  nr = nrz*nrx
  if (exist_file("receivers_zxt")) then
    call from_aux("receivers_zxt","n2",nr,nr)
  end if
  allocate(receiver_zxt(nr,4));  receiver_zxt = 0;
  if (exist_file("receiver_zxt")) then
    allocate(receiver_zxt_in(nr,4)); receiver_zxt_in=0.
    call sreed("receiver_zxt",receiver_zxt_in,4*size(receiver_zxt_in))
    do ir=1,nr
      receiver_zxt(ir,:) = (/ 1+nint((receiver_zxt_in(ir,1)-z0)/dz), &
                              1+nint((receiver_zxt_in(ir,2)-x0)/dx), &
                              1+nint( receiver_zxt_in(ir,3)/dt_comp), int(receiver_zxt_in(ir,4)) /)
    end do
    deallocate(receiver_zxt_in)
  else
    do irx=1,nrx
    do irz=1,nrz
      receiver_zxt((irx-1)*nrz+irz,:) = (/ 1+frz+(irz-1)*jrz,1+frx+(irx-1)*jrx,1,rtype /)
    end do
    end do
  end if
  
  call from_param("dt_data",dt_data,.001)
  jt_data = int(dt_data/dt_comp)
  dt_data = dt_comp*jt_data      ! After rounding
  if (nrt.eq.-1) then; nrt = int(T/dt_data); end if
  nt_data = nrt
  allocate(data(nt_data,nr)); data=0.;
  
  ! Create output header for data
  call to_history("n1",nt_data)
  call to_history("n2",nr)
  call to_history("n3",1 )
  
  call to_history("d1",dt_data)
  call to_history("d2",dx*jrx)
  call to_history("d3",1.)
  
  call to_history("o1",0.)
  call to_history("o2",x0+frx*dx*jrx)
  call to_history("o3",0.)
  
  
  !! Snapshot output
  if (exist_file("snap")) then
    snap = .true.
    call from_param("dt_snap",dt_snap,T  )
    jt_snap = int(dt_snap/dt_comp)
    dt_snap = dt_comp*jt_snap      ! After rounding
    nt_snap = int(T/dt_snap)  

    ! Create output header for snapshots
    call to_history("n1",nz,"snap")
    call to_history("n2",nx,"snap")
    call to_history("n3",nt_snap,"snap")
    
    call to_history("d1",dz,"snap")
    call to_history("d2",dx,"snap")
    call to_history("d3",dt_snap,"snap")
    
    call to_history("o1",z0,"snap")
    call to_history("o2",x0,"snap")
    call to_history("o3",0.,"snap")
    
    allocate(usnap (nz,nx,nt_snap)); usnap=0.;
  else
    snap = .false.
  end if


  !! Sources Input grid
  if (exist_file("sources_zxt")) then
    call from_aux("sources_zxt","n1",ns)
    allocate(sources_zxt   (ns,4)); sources_zxt = 0.
    allocate(sources_zxt_in(ns,4)); sources_zxt_in = 0.
    call sreed("sources_zxt",sources_zxt_in,4*size(sources_zxt_in))
    do is=1,ns
      sources_zxt(is,:) = (/ 1+nint((sources_zxt_in(is,1)-z0)/dz), &
                             1+nint((sources_zxt_in(is,2)-x0)/dx), &
                             1+nint( sources_zxt_in(is,3)/dt_comp), int(sources_zxt_in(is,4)) /)
    end do
    deallocate(sources_zxt_in)
  else
    call from_param("nsz",nsz,1 )
    call from_param("nsx",nsx,1 )
    call from_param("fsz",fsz,nz/2+1)
    call from_param("fsx",fsx,nx/2+1)
    call from_param("jsz",jsz,1 )
    call from_param("jsx",jsx,1 )
    call from_param("stype",stype,1)
    ns = nsz*nsx
    allocate(sources_zxt   (ns,4)); sources_zxt = 0.
    do isx=1,nsx
    do isz=1,nsz
      sources_zxt((isx-1)*nsz+isz,:) = (/ 1+fsz+(isz-1)*jsz,1+fsx+(isx-1)*jsx,1,stype /)
    end do
    end do
  end if

  !! Sources Functions
  call from_param("smth",smth,0) ! Source smoothing over smth gridpoints
  ! Prepare smoothing weighting factor
  smth_fact = 0.   ! Weighting factor because of source smoothing
  do ix=-smth,smth; do iz=-smth,smth
      smth_fact = smth_fact + .5*(cos(sqrt(float((ix)**2+(iz)**2))/3.*pi)+1.)
  end do;           end do

  if (exist_file("sources")) then
   ! Read source functions from file
    call from_aux("sources","n1",nst)
    allocate(sources(-nst:nst,ns));  sources = 0.
    call sreed("sources",sources(1:nst,ns),4*nst*ns)
  else
    call from_param("nst",nst,128 )
    allocate(sources(-nst:nst,ns));  sources = 0.
    call from_param("f0",f0,25.); w0 = 2.*pi*f0
    call from_param("ssquare",ssquare,.false.)
    ! Construct Ricker source wavelet
    do is=1,ns
      call sourcericker(sources(-nst:nst,is),2*nst,dt_comp,w0,1,ssquare)
      sources(:,is) = sources(:,is)*dt_comp*dt_comp*vv(sources_zxt(is,1),sources_zxt(is,2))**2./(dx*dz)/smth_fact
    end do
  end if
  
  ! Prepare injection grid (based on source type an smoothing)

  ! Complicated source injection grid count Count
  ns2=0
  do is=1,ns
     ! Classic right-hand side source
     if      (sources_zxt(is,4).eq.0) then
       do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
          do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
             ns2=ns2+1
          end do
       end do
         
     ! Impulsive volume injection-rate source type ( wavelet dt already done by sourcericker() )
     else if (sources_zxt(is,4).eq.1) then
        do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
           do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
              ! +
              ns2=ns2+1
              ! -
              ns2=ns2+1
           end do
        end do

     ! Impulsive force source source type (x-direction)
     else if (sources_zxt(is,4).eq.2) then
        do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
           do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
              ! +
              ns2=ns2+1
              ! -
              ns2=ns2+1
           end do
        end do

     ! Impulsive force source source type (z-direction)
     else if (sources_zxt(is,4).eq.3) then
        do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
           do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
              ! +
              ns2=ns2+1
              ! -
              ns2=ns2+1
           end do
        end do
      end if
  end do
  
  
  allocate(us   (-nst:nt_comp,ns2  )); us    = 0.
  allocate(us_zx(             ns2,2)); us_zx = 0
  
  
  it=-nst+2
  do while (it.le.nt_comp)
    is2 = 1
    do is=1,ns

         ! Classic right-hand side source
         if      (sources_zxt(is,4).eq.0) then
           do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
              do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
                 rr=sqrt(float((ix-sources_zxt(is,2))**2+(iz-sources_zxt(is,1))**2))
                 if(it-sources_zxt(is,3)+1.le.nst.and.it-sources_zxt(is,3)-1.ge.-nst)then
                   us   (it,is2) = sources(it-sources_zxt(is,3),is)*.5*(cos(rr/3*pi)+1.)
                 else
                   us   (it,is2) = 0.
                 end if
                 us_zx(is2, :) = (/ iz, ix /)
                 is2           = is2 + 1
              end do
           end do
            
         ! Impulsive volume injection-rate source type
         else if (sources_zxt(is,4).eq.1) then
           do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
              do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
                 rr=sqrt(float((ix-sources_zxt(is,2))**2+(iz-sources_zxt(is,1))**2))
                 ! +
                 if(it-sources_zxt(is,3)+1.le.nst.and.it-sources_zxt(is,3)-1.ge.-nst)then
                   us   (it,is2) = rh(iz,ix)/(2*dt_comp)*sources(it-sources_zxt(is,3)+1,is)*.5*(cos(rr/3*pi)+1.)
                 else
                   us   (it,is2) = 0.
                 end if
                 us_zx(is2, :) = (/ iz, ix /)
                 is2           = is2 + 1
                 ! -
                 if(it-sources_zxt(is,3)+1.le.nst.and.it-sources_zxt(is,3)-1.ge.-nst)then
                   us   (it,is2) = rh(iz,ix)/(2*dt_comp)*sources(it-sources_zxt(is,3)-1,is)*.5*(cos(rr/3*pi)+1.)
                 else
                   us   (it,is2) = 0.
                 end if
                 us_zx(is2, :) = (/ iz, ix /)
                 is2           = is2 + 1
              end do
           end do

         ! Impulsive force source source type (x-direction)
         else if (sources_zxt(is,4).eq.2) then
           do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
              do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
                 ! +
                 rr=sqrt(float((ix+1-sources_zxt(is,2))**2+(iz-sources_zxt(is,1))**2))
                 if(it-sources_zxt(is,3)+1.le.nst.and.it-sources_zxt(is,3)-1.ge.-nst)then
                   us   (it,is2) = -1./(2*dx)*sources(it-sources_zxt(is,3),is)*.5*(cos(rr/3*pi)+1.)
                 else
                   us   (it,is2) = 0.
                 end if
                 us_zx(is2, :) = (/ iz, ix /)
                 is2           = is2 + 1
                 ! -
                 rr=sqrt(float((ix-1-sources_zxt(is,2))**2+(iz-sources_zxt(is,1))**2))
                 if(it-sources_zxt(is,3)+1.le.nst.and.it-sources_zxt(is,3)-1.ge.-nst)then
                   us   (it,is2) = +1./(2*dx)*sources(it-sources_zxt(is,3),is)*.5*(cos(rr/3*pi)+1.)
                 else
                   us   (it,is2) = 0.
                 end if
                 us_zx(is2, :) = (/ iz, ix /)
                 is2           = is2 + 1
              end do
           end do

         ! Impulsive force source source type (z-direction)
         else if (sources_zxt(is,4).eq.3) then
           do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
              do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
                 ! +
                 rr=sqrt(float((ix-sources_zxt(is,2))**2+(iz+1-sources_zxt(is,1))**2))
                 if(it-sources_zxt(is,3)+1.le.nst.and.it-sources_zxt(is,3)-1.ge.-nst)then
                   us   (it,is2) = -1./(2*dz)*sources(it-sources_zxt(is,3),is)*.5*(cos(rr/3*pi)+1.)
                 else
                   us   (it,is2) = 0.
                 end if
                 us_zx(is2, :) = (/ iz, ix /)
                 is2           = is2 + 1
                 ! -
                 rr=sqrt(float((ix-sources_zxt(is,2))**2+(iz-1-sources_zxt(is,1))**2))
                 if(it-sources_zxt(is,3)+1.le.nst.and.it-sources_zxt(is,3)-1.ge.-nst)then
                   us   (it,is2) = +1./(2*dz)*sources(it-sources_zxt(is,3),is)*.5*(cos(rr/3*pi)+1.)
                 else
                   us   (it,is2) = 0.
                 end if
                 us_zx(is2, :) = (/ iz, ix /)
                 is2           = is2 + 1
              end do
           end do
         end if

    end do

    ! Update time counter
    it = it+1

  end do
  
  
  
  ! Calculate number of recording positions
  nr2 = 0
  do ir = 1,nr
    
    ! Record Pressure
    if      (receiver_zxt(ir,4).eq.1) then
      nr2 = nr2+1
    ! Record Vx
    else if (receiver_zxt(ir,4).eq.2) then
      nr2 = nr2+1
      nr2 = nr2+1
    ! Record Vz
    else if (receiver_zxt(ir,4).eq.3) then
      nr2 = nr2+1
      nr2 = nr2+1
    end if
    
  end do
  
  allocate(ur   (-nst:nt_comp,nr2  )); ur    = 0.
  allocate(ur_zx(             nr2,2)); ur_zx = 0
  
  ir2 = 1
  do ir = 1,nr
    
    ! Record Pressure
    if      (receiver_zxt(ir,4).eq.1) then
       ur_zx(ir2,:) = (/ receiver_zxt(ir,1) , receiver_zxt(ir,2)  /)
       ir2 = ir2+1
    
    ! Record Vx
    else if (receiver_zxt(ir,4).eq.2) then
       ur_zx(ir2,:) = (/ receiver_zxt(ir,1) , receiver_zxt(ir,2)+1 /)
       ir2 = ir2+1
       ur_zx(ir2,:) = (/ receiver_zxt(ir,1) , receiver_zxt(ir,2)-1 /)
       ir2 = ir2+1
    
    ! Record Vz
    else if (receiver_zxt(ir,4).eq.3) then
       ur_zx(ir2,:) = (/ receiver_zxt(ir,1)+1 , receiver_zxt(ir,2) /)
       ir2 = ir2+1
       ur_zx(ir2,:) = (/ receiver_zxt(ir,1)-1 , receiver_zxt(ir,2) /)
       ir2 = ir2+1
    end if

  end do
  
 ! write(0,*) 'ur_zx(:,1)'
 ! write(0,*)  ur_zx(:,1)
 ! write(0,*) 'ur_zx(:,2)'
 ! write(0,*)  ur_zx(:,2)

 ! write(0,*) 'nr2'
 ! write(0,*)  nr2


  call sep_close()
  
  
  
  
  
  ! *********************************************************************
  !  Setup Finite Differences
  ! *********************************************************************
  write(0,*) '----------------------------'
  write(0,*) 'Setup Finite Differences'
  
  !allocate space for the velocity model, the absorbing boundaries and the wavefields
  allocate(u (1-nop:nz+nop,1-nop:nx+nop,3)); u  = 0.
  allocate(ut(1-nop:nz+nop,1-nop:nx+nop)  ); ut = 0.
  a1x = 0.
  a2x = 0.
  a3x = 0.
  a4x = 0.
  a5x = 0.
  a1z = 0.
  a2z = 0.
  a3z = 0.
  a4z = 0.
  a5z = 0.
  
  ! call the setup routine to build a propagator for the main grid
  call setupfd(a1x,a2x,a3x,a4x,a1z,a2z,a3z,a4z,dx,dz)
  
  ! Prepare Taper
  allocate(taper(npad+nop))
  do iz=1,npad+nop
     taper(iz) = exp(-tapco*iz*iz)
  end do
  
  



  
  call stop_timer_num (tm2)
  call start_timer_num(tm3)
  ! *********************************************************************
  !  Loop over the total time
  ! *********************************************************************
  write(0,*) '----------------------------'
  write(0,*) 'Loop over the total time'

  it=-nst+2
  id=1
  in=1
  
  do while (it.le.nt_comp)
    
    if (freesurf) then
  !     !! anti-symetrize the  p-wave potential above the free surface
       do ix=1,nx
             u( 0,ix,2) =  0.
             u(-1,ix,2) =  0.
             u(-2,ix,2) =  0.
             u(-3,ix,2) =  0.
             u(-4,ix,2) =  0.
       end do
    end if
    

    ! Apply laplacian

    do iz=1,nz  ! backward derivative
      ut(iz,:) = a1z * (u(iz+0,:,2) - u(iz-1,:,2)) + &
                 a2z * (u(iz+1,:,2) - u(iz-2,:,2)) + &
                 a3z * (u(iz+2,:,2) - u(iz-3,:,2)) + &
                 a4z * (u(iz+3,:,2) - u(iz-4,:,2))
    end do
    do ix=1,nx  ! multiply with inverse of density
      do iz=1,nz
        ut(iz,ix) = ut(iz,ix)/rh(iz,ix)
      end do
    end do
    do iz=1,nz   ! forward derivative
      u(iz,:,3) = a1z * (ut(iz+1,:) - ut(iz-0,:)) + &
                  a2z * (ut(iz+2,:) - ut(iz-1,:)) + &
                  a3z * (ut(iz+3,:) - ut(iz-2,:)) + &
                  a4z * (ut(iz+4,:) - ut(iz-3,:))
    end do

    
    
    do ix=1,nx   ! forward derivative
      ut(:,ix) = a1x * (u(:,ix+0,2) - u(:,ix-1,2)) + &
                 a2x * (u(:,ix+1,2) - u(:,ix-2,2)) + &
                 a3x * (u(:,ix+2,2) - u(:,ix-3,2)) + &
                 a4x * (u(:,ix+3,2) - u(:,ix-4,2))
    end do
    do ix=1,nx  ! multiply with inverse of density
      do iz=1,nz
        ut(iz,ix) = ut(iz,ix)/rh(iz,ix)
      end do
    end do
    do ix=1,nx  ! backward derivative
      u(:,ix,3) = u(:,ix,3) + &
                  a1x * (ut(:,ix+1) - ut(:,ix-0)) + &
                  a2x * (ut(:,ix+2) - ut(:,ix-1)) + &
                  a3x * (ut(:,ix+3) - ut(:,ix-2)) + &
                  a4x * (ut(:,ix+4) - ut(:,ix-3))
    end do

    ! 'Add' the time derivative and multiply with inverse of incompressibility
    do ix=1,nx
      do iz=1,nz
         u(iz,ix,3) = dt_comp*dt_comp*ik(iz,ix)*u(iz,ix,3)
      end do
    end do
    
   
    !! taper
    if (.not.freesurf) then
    do ix=1-nop,nx+nop
      do iop=1,npad+nop
      u( 1+npad-iop,ix,1) = u( 1+npad-iop,ix,1) * taper(iop)
      u( 1+npad-iop,ix,2) = u( 1+npad-iop,ix,2) * taper(iop)
      u( 1+npad-iop,ix,3) = u( 1+npad-iop,ix,3) * taper(iop)
      end do
    end do
    end if
    do iz=1-nop,nz+nop
      do iop=1,npad+nop
      u(iz, 1+npad-iop,1) = u(iz, 1+npad-iop,1) * taper(iop)
      u(iz, 1+npad-iop,2) = u(iz, 1+npad-iop,2) * taper(iop) 
      u(iz, 1+npad-iop,3) = u(iz, 1+npad-iop,3) * taper(iop)
      end do
    end do
    do iz=1-nop,nz+nop
      do iop=1,npad+nop
      u(iz,nx-npad+iop,1) = u(iz,nx-npad+iop,1) * taper(iop)
      u(iz,nx-npad+iop,2) = u(iz,nx-npad+iop,2) * taper(iop)
      u(iz,nx-npad+iop,3) = u(iz,nx-npad+iop,3) * taper(iop)
      end do
    end do
   
    do ix=1-nop,nx+nop
      do iop=1,npad+nop
      u(nz-npad+iop,ix,1) = u(nz-npad+iop,ix,1) * taper(iop)
      u(nz-npad+iop,ix,2) = u(nz-npad+iop,ix,2) * taper(iop)
      u(nz-npad+iop,ix,3) = u(nz-npad+iop,ix,3) * taper(iop)
      end do
    end do
   
    ! 'Add' the time derivative and multiply with inverse of incompressibility
    do ix=1,nx
      do iz=1,nz
         u(iz,ix,3) = u(iz,ix,3) + 2.*u(iz,ix,2) - u(iz,ix,1)
      end do
    end do

   
   
    !! Add the source wavelet at the source locations
    do is2=1,ns2
      u(us_zx(is2,1),us_zx(is2,2),3)=u(us_zx(is2,1),us_zx(is2,2),3) + us(it,is2)
    end do
    

    ! Save the wavefield for recording
    do ir2=1,nr2
      ur(it,ir2) = u(ur_zx(ir2,1),ur_zx(ir2,2),3)
    end do


    

    !! Now output the desired snapshots if they were requested  
    !!   they are calculated at only every jt_snap time steps
    if ( snap ) then
      if ( mod(it,jt_snap).eq.0 .and. it.ge.0 .and. in.le.nt_snap ) then
           do ix=1,nx
           do iz=1,nz
              usnap(iz,ix,in)=u(iz,ix,3)
           end do
           end do
        in=in+1
      end if
    end if


    !!! CYCLE
    ! Update  all the variables in time
    do ix=1,nx
        do iz=1,nz
           u(iz,ix,1)=u(iz,ix,2)
           u(iz,ix,2)=u(iz,ix,3)
        end do
    end do


    ! Time step verbose
    if (mod(it-1,jt_verb).eq.0.) then
      write(0,*) "Step:",nint(it*dt_comp),'seconds'
    end if
    
    ! Update time counter
    it = it+1
    
  end do
  ! *********************************************************************
  !  End of Loop over the total time
  ! *********************************************************************
  


  
  it=-nst+2
  do while (it.le.nt_comp)
    ir2 = 1
    
    !! Save the current wavefield at the geophones as our seismogram: Save the
    !! locations that were requested only.
    if ( mod(it,jt_data).eq.0 .and. it.ge.0 .and. id.le.nt_data ) then
      do ir=1,nr
        
        ! Record Pressure
        if      (receiver_zxt(ir,4).eq.1) then
          data(id,ir) = ur(it,ir2)
          ir2 = ir2 + 1
          
        ! Record Vx
        else if (receiver_zxt(ir,4).eq.2) then
          data(id,ir) = - ( ur(it-1,ir2) - ur(it-1,ir2+1) )/(rh(iz,ix)*2*dx)*dt_comp  &
                        - ( ur(it  ,ir2) - ur(it  ,ir2+1) )/(rh(iz,ix)*2*dx)*dt_comp
          ir2 = ir2 + 1
          
        ! Record Vz
        else if (receiver_zxt(ir,4).eq.3) then
          data(id,ir) = - ( ur(it-1,ir2) - ur(it-1,ir2+1) )/(rh(iz,ix)*2*dx)*dt_comp  &
                        - ( ur(it  ,ir2) - ur(it  ,ir2+1) )/(rh(iz,ix)*2*dx)*dt_comp
          ir2 = ir2 + 1
          
        end if
        
      end do
      id=id+1
    end if

    ! Update time counter
    it = it+1
    
  end do
  
  
  ! Write out data and snaps
  call srite("out" ,data ,4*size(data ))
  if ( snap ) then
    call srite("snap",usnap,4*size(usnap))
  end if

  ! Final
  write(0,*) '----------------------------'
  call stop_timer_num (tm3)
  call stop_timer_num (tm1)
  call print_timers()


  deallocate(vv )
  deallocate(rh )
  deallocate(u  )
  deallocate(receiver_zxt)
  deallocate(sources_zxt)
  deallocate(data)
  deallocate(sources)
  if ( snap ) then
    deallocate(usnap)
  end if
  deallocate(taper)
  deallocate(ur)
  deallocate(ur_zx)
  deallocate(us)
  deallocate(us_zx)

  write(0,*) '----------------------------'
  call exit(0)

end program





  subroutine setupfd(a1x,a2x,a3x,a4x,a1z,a2z,a3z,a4z,dx,dz)

    implicit none
    real    :: dx,dz
    real    :: a1x,a2x,a3x,a4x
    real    :: a1z,a2z,a3z,a4z


    ! derivative coefficients (Paul Sava's)
    a1x= 1.      *1225./1024./dx
    a2x=-1./15.  *1225./1024./dx
    a3x= 1./125. *1225./1024./dx
    a4x=-1./1715.*1225./1024./dx
    a1z= 1.      *1225./1024./dz
    a2z=-1./15.  *1225./1024./dz
    a3z= 1./125. *1225./1024./dz
    a4z=-1./1715.*1225./1024./dz

    return

  end subroutine setupfd


