module fop_mod
  ! Weighting operator
  
  implicit none
  
  integer,private                             :: nn
  double precision, allocatable               :: f(:)
  public :: fop_init, fop_close, fop_lop, fop_op
  
  contains
  
  subroutine fop_init( w , n )
    double precision, optional  :: w(:)
    integer,optional            :: n
    call fop_close()
    if      (present(n)) then; nn=n
    else if (present(w)) then; nn = size(w);
    else; write(0,*) "No aruments supplied to fop_init()"; call exit(0); end if
    allocate(f(nn)); f = 1.
    if (present(w))  f = w
  end subroutine
  
  subroutine fop_close()
    if (allocated(f)) deallocate(f)
  end subroutine
  
  
  
  ! Generic operator for SEP solvers
  function fop_lop(adj,add,m,d) result(stat)
    integer             :: stat
    logical,intent(in)  :: adj , add
    double precision,intent(inout)  :: m(:), d(:)
    call adj_null(adj,add,m,d)
    if (.not.adj) call fop_for(m,d)
    if (     adj) call fop_adj(m,d)
    stat=0
  end function
  
  subroutine fop_op(adj,add,m,d)
    integer             :: stat
    logical,intent(in)  :: adj , add
    double precision,intent(inout)  :: m(:), d(:)
    call adj_null(adj,add,m,d)
    if (.not.adj) call fop_for(m,d)
    if (     adj) call fop_adj(m,d)
  end subroutine
  



  ! Forward operator  
  subroutine fop_for( m, d)
    double precision,intent(in)     :: m(:)
    double precision,intent(out)    :: d(:)
    integer             :: ii
    do ii=1,nn
      d(ii) = d(ii) + f(ii)*m(ii)
    end do
  end subroutine

  ! Adjoint operator
  subroutine fop_adj( m, d)
    double precision,intent(out)    :: m(:)
    double precision,intent(in)     :: d(:)
    integer             :: ii
    do ii=1,nn
      m(ii) = m(ii) + f(ii)*d(ii)
    end do
  end subroutine


  
  
  subroutine     adj_null(adj,add,x,y)
    logical,intent(in) :: adj , add
    double precision   :: x(:), y(:)
    if (.not.add) then
      if (     adj)  x = 0.
      if (.not.adj)  y = 0.
    end if
  end subroutine


end module 
