module gradiometry_reg_mod
  ! Contains tho operators needed for gradiometry on a regular grid
  implicit none

  private
  public :: gradiometry_reg_init, gradiometry_reg_close, grad_laplacian_reg, grad_deltdelt
  integer :: n1,n2
  real    :: d1,d2,dt
  
  contains
  
  subroutine gradiometry_reg_init(n1_in,n2_in,d1_in,d2_in,dt_in)
    integer             :: n1_in,n2_in
    real                :: d1_in,d2_in,dt_in
    n1=n1_in
    n2=n2_in
    d1=d1_in
    d2=d2_in
    dt=dt_in
  end subroutine
  
  subroutine gradiometry_reg_close()
  end subroutine
  
  
  ! Laplace operator  
  subroutine grad_laplacian_reg(m,d)
    real,intent(in)                 :: m(n1,n2)
    double precision,intent(out)    :: d(n1,n2)
    integer                         :: ii,jj
    
    d=0.
    do ii=2,n1-1
    do jj=2,n2-1
      d(ii,jj) = m(ii-1,jj)+m(ii,jj-1)+m(ii+1,jj)+m(ii,jj+1)-4*m(ii,jj)
    end do
    end do
    
    jj=1
    do ii=2,n1-1
      d(ii,jj)=d(ii,jj+1)
    end do
    jj=n2
    do ii=2,n1-1
      d(ii,jj)=d(ii,jj-1)
    end do

    ii=1
    do jj=2,n2-1
      d(ii,jj)=d(ii+1,jj)
    end do
    ii=n1
    do jj=2,n2-1
      d(ii,jj)=d(ii-1,jj)
    end do
    
    d( 1, 1)=d(   2,   2)
    d(n1, 1)=d(n1-1,   2)
    d(n1,n2)=d(n1-1,n2-1)
    d( 1,n2)=d(   2,n2-1)

    d = d/d1/d2

  end subroutine
  
  
  ! Second order time derivative  
  subroutine grad_deltdelt(m,d)
    real,intent(in)                 :: m(n1,n2,3)
    double precision,intent(out)    :: d(n1,n2)
    integer             :: ii,jj

    d=0.
    do ii=1,n1
    do jj=1,n2
      d(ii,jj) = m(ii,jj,1)+m(ii,jj,3)-2*m(ii,jj,2)
    end do
    end do
    
    d = d/dt/dt
    
  end subroutine



end module
