program rickerwavelet
  
  ! nt: nr of points in full fft
  ! n1: nr of timelags to write-out
  ! sym_t: give positive and negative timelags
  
  ! f0: center freq
  ! t0: shifted t
  
  
  use sep
  use fftw_mod
  use fftshift_mod
  use fftshift_io_mod
  implicit none
  
  
  integer              ::  n1,nt,nw
  real                 ::  o1,dt, w, w0, f0, dw,t0,ow
  real, parameter      ::  pi=3.14159265
  
  integer              :: iw

  logical              :: sym_t

  real,allocatable     :: datat(:)
  complex,allocatable  :: dataw(:)
  
  
  ! Input&output
  call sep_init()
  
  call from_param("nt",nt,128 )
  call from_param("n1",n1,nt/2)
  call from_param("d1",dt,.001)
  call from_param("o1",o1,.0  )
  
  call from_param("sym_t",sym_t,.true.)

  call from_param("f0",f0,25.); w0 = 2.*pi*f0
  call from_param("t0",t0, 0.)

  if (sym_t) then;  call to_history("n1",2*n1);
  else;             call to_history("n1",  n1); end if
  call to_history("dt",dt);
  if (sym_t) then;  call to_history("o1",-n1*dt);
  else;             call to_history("o1",    o1); end if

  call sep_close()
  
  
  ! Calculate frequency units 
  nw =       nt/2+1
  dw = 2.*pi/(nt*dt)
  ow =        0.
  
  allocate(dataw(nw),datat(nt))
  dataw=0.; datat=0.
  
  ! Init IFFT
  call fftw1d_1_r2c_lop_init(nt_in=nt)
  
  ! Construct frequency domain wavelet
  do iw = 1,nw
    w   = ow+(iw-1)*dw
    dataw(iw) = ( (2.*w**2.) / (sqrt(pi)*w0**3.) ) * exp(-w**2./w0**2.)*cexp(cmplx(0.,-w*t0))
  end do

  ! Do IFFT
  call fftw1d_1_r2c_lop(.true.,.false.,datat,dataw)
  datat = 2.*datat

  ! Write data
  if (sym_t) then;  call fftshift1d_srite("out" ,datat      ,2*n1,4);
  else;             call            srite("out", datat(1:n1),  n1*4);  end if
  
  ! Close, deallocate and exit
  call fftw1d_1_r2c_lop_close()
  deallocate(dataw,datat)
  call exit()

end program
