module solver_reg_mod 
  ! 0 = W (F m - d)
  ! 0 =    A m
  use solver_reg_chain_mod
  use solver_report_mod
  implicit none
  
  logical, parameter, private  :: T = .true., F = .false.
  contains
  
  subroutine solver_reg( m,d, Fop, Aop, stepper, nAop, niter,eps0,  &
    &                    Wop,m0,rm0,err,resd,resm,mmov,rmov,verb)
    optional          :: Wop,m0,rm0,err,resd,resm,mmov,rmov
    interface  
      integer function Fop(adj,add,m,d)
        double precision::m(:),d(:)
        logical,intent(in)::adj,add
      end function 
      integer function Aop(adj,add,m,d)
        double precision::m(:),d(:)
        logical,intent(in)::adj,add
      end function 
      integer function Wop(adj,add,m,d)
        double precision::m(:),d(:)
        logical,intent(in)::adj,add
      end function 
      integer function stepper(forget,m,g,rr,gg)
       double precision, dimension(:)  ::        m,g,rr,gg    
        logical             :: forget
      end function
    end interface 
    double precision,dimension(:),    intent(in)          :: d, m0,rm0
    integer,               intent(in)          :: niter, nAop
    logical,               intent(in)          :: verb
    double precision,                  intent(in)          :: eps0
    double precision                                       :: eps
    double precision, dimension(:),    intent(out)         :: m,    resd,resm
    double precision, dimension(:,:),  intent(out)         ::   err,rmov,mmov
    double precision, dimension(size( m))                  :: g
    double precision, dimension(size( d) + nAop), target   :: rr, gg, tt
    double precision, dimension(:), pointer                :: rd, gd, td
    double precision, dimension(:), pointer                :: rm, gm, tm
    integer                                    :: iter, stat
    logical                                    :: forget

    ! Direct pointers
    rd => rr(1:size(d))
    rm => rr(1+size(d):) ;  
    gd => gg(1:size(d))
    gm => gg(1+size(d):)
    td => tt(1:size(d))
    tm => tt(1+size(d):)
    
    
    ! Zero model and model-residual
    m  = 0.
    rm = 0.
    
    ! begin initialization
    eps=eps0
    if (present(rm0)) rm = rm0             ! Rm  =    Rm0
    if (present( m0))  m =  m0             ! m   =     m0
    if (present(Wop)) then
      stat=       Wop(F,F,-d,rd)           ! Rd  = -W   d
      call chain0(Wop,Fop,F,T,m,rd,td)     ! Rd +=  WF m0
    else
      rd   =              -d               ! Rd  = -    d
      stat =          Fop(F,T,m,rd   )     ! Rd +=   F m0
    end if 
      stat = Aop(F,T,eps*m,rm)             ! Rm += e A m0                      
    forget = T
    
    ! Begin iterations
    do iter = 1,niter
      if(verb) write(0,*) "iter",iter
      if (present( Wop)) then
        call chain0(Wop,Fop,T,F,g,rd,td)   ! g   =  (WF)'Rd
      else
        stat =          Fop(T,F,g,rd   )   ! g   =    F' Rd
      end if 
      stat =  Aop(T,T,g,eps*rm)            ! g  +=  e A' Rm                       
      if (present( Wop)) then
        call chain0(Wop,Fop,F,F,g,gd,td)   ! Gd  =  (WF) g
      else
        stat =          Fop(F,F,g,gd   )   ! Gd  =    F  g
      end if 
      stat =  Aop(F,F,eps*g,gm)            ! Gm  =  e A  g                         
      
      ! m += dm; R += dR
      stat = stepper(forget, m,g, rr,gg)               
      if (stat.eq.1) exit ! got stuck descending
      if (present( mmov)) mmov(:,iter) =  m(:size(mmov,1))
      if (present( rmov)) rmov(:,iter) = rr(:size(rmov,1))
      if (present( err)) then
        err(iter,1) = dot_product(rd,rd)
        err(iter,2) = dot_product(gd,gd)
        err(iter,3) = dot_product(rm,rm)
        err(iter,4) = dot_product(gm,gm)
      end if
   !   if (verb) call solver_report(iter,m,g,rd,rm)
      forget=F            ! Steepest descent ? (think not)
    end do
    
    if (present( resd)) resd = rd
    if (present( resm)) resm = rm(:size(resm))
    
  end subroutine
  
end module 
