module sources_mod
  
   implicit none
  
   contains

   subroutine sourcefn(source,nts,fmin,fmax,dt)

   
   integer ::nts,nf,it,if

   real    ::fmin,fmax,dt,tee,pi,freq

   real,dimension(:),allocatable  ::taper

   real    ::source(-nts:nts)

   allocate(taper(-nts:nts)); taper=0;

   pi=3.1415926535
   nf=100

   do it=-nts,nts
      taper(it)=.5+.5*cos(float(it)*pi/float(nts))
   end do

   source=0.

   do if=1,nf

      freq=float(if-1)/float(nf)*(fmax-fmin)+fmin

      do it=-nts,nts
         tee=it*dt
         source(it)=source(it)+cos(2.*pi*freq*tee)*taper(it)     
      end do
   end do

   deallocate(taper)
   return

   end subroutine sourcefn



   subroutine sourcericker(source,nt,dt,w0,it0,ssquare)
   
   use fftw_mod
   use fftshift_mod


   integer                          ::  nt,nw,it,iw,it0
   real                             ::  fmin,fmax,dt,tee,freq,w,dw,ow,w0,fact
   real,dimension(:),allocatable    ::  datat
   complex,dimension(:),allocatable ::  dataw
   real                             ::  source(nt)
   real,parameter                   ::  pi=3.1415926535
   logical                          ::  ssquare
   
   ! Calculate frequency units 
   nw =    nt/2+1
   dw = 2.*pi/(nt*dt)
   ow =        0.
   fact = 2./sqrt(real(nt))/dt

   allocate(dataw(nw),datat(nt))
   dataw=0.; datat=0.

   ! Init IFFT
   call fftw1d_1_r2c_lop_init(nt_in=nt)
   
   ! Construct frequency domain wavelet
   do iw = 1,nw
     w   = ow+(iw-1)*dw
     dataw(iw) = ( (2.*w**2.) / (sqrt(pi)*w0**3.) ) * exp(-w**2./w0**2.)*cexp(cmplx(0.,-w*(it0-1)*dt))
   end do

   ! Optionally square the wavelet
   if (ssquare) then
     dataw = dataw**2.
   end if

   ! Do IFFT
   call fftw1d_1_r2c_lop(.true.,.false.,datat ,dataw)
   call fftshift1d_1_lop(.true.,.false.,source,datat)

   ! Taper and multiply by 2*sqrt(2pidt/df)
   do it=1,nt
     source(it) = source(it)*fact
   end do

   ! Close, deallocate and exit
   call fftw1d_1_r2c_lop_close()
   deallocate(dataw,datat)
   return
   
   end subroutine sourcericker











  subroutine rndsourcepos(source_zxt,ns,sz,nz,nx,nt,dist1,dist2,delzs,gz,gx)

    integer  :: ns,sz,delzs,is,nz,nx,nt,dist1,dist2
    real     :: gx,gz,harvest_x,harvest_z,harvest_t,harvest_e,expectency
    integer     :: source_zxt(3,ns)

    is=0
    do while (is.lt.ns)

      call random_number(harvest_x);  harvest_x = 1+aint(harvest_x*nx)
      call random_number(harvest_z);  harvest_z = 1+aint( sz+(harvest_z-.5)*2*delzs) 
      call random_number(harvest_t);  harvest_t = 1+aint(harvest_t*nt) 
      call random_number(harvest_e);
      if (dist1.eq.1) then
      if (dist2.eq.1) then
        expectency = 1.
      else
        expectency = 1.
      end if
      else
      if (dist2.eq.1) then
        expectency = 1.
      else
        expectency = 1.
      end if
      end if
      
      if (harvest_e.le.expectency) then
        is = is+1
        source_zxt(:,is)=(/ harvest_z,harvest_x,harvest_t /)
      end if

    end do


  end subroutine



  
  
end module
